% -*- MATLAB -*-
% Time-stamp: "2015-11-08 11:39:03 user"
%----------------------------------------------------------------------
%  Copyright (C) 2015, FUJIWARA,Hiroshi
%    <fujiwara@acs.i.kyoto-u.ac.jp>
%----------------------------------------------------------------------

classdef exfloat

  properties (Constant, Access=private)

    %----------------------------------------
    % required precision in decimal digits
    %----------------------------------------
    precision10 = 500;

    % precision in 2^64 radix, log(10)/log(2)/64 = 0.051905126
    precision64 = int32(exfloat.precision10 * 0.051905126) + 1;

    % actual arithmetic precision, 64*log(2) = 19.26591972
    precision = 19.26591972 * double(exfloat.precision64);

    % sum of exponent part and fraction part
    size64 = int32(1 + exfloat.precision64);
    byte   = int32(8 * exfloat.size64);

    % default output format
    default_output_precision = int32(4);
    default_output_format    = 'd';         % 'd' means 'as double'
    default_display_format   = '%.20g';
    default_fixed_scientific = -4;

    SIGN     = uint64( hex2dec('8000000000000000') );
    NEGINFTY = uint64( hex2dec('ffffffffffffffff') );
    F_FORMAT = uint64( hex2dec('0400000000000000') );
    E_FORMAT = uint64( hex2dec('0200000000000000') );
    EXP_MASK = bitshift(uint64(hex2dec('7fffffff')), 32)+uint64(hex2dec('ffffffff'));
    BIAS     = bitshift(uint64(hex2dec('3fffffff')), 32)+uint64(hex2dec('ffffffff'));

  end

  %properties (Access=private)
  properties (Access=public)
    num
  end

  methods

    %------------------------------
    % constructor
    %------------------------------
    function x = exfloat( a )

      if nargin == 0
        x.num = zeros(exfloat.size64,1, 'uint64');
        return;
      end
      
      [row,col] = size( a );

      if isa(a, 'exfloat')

        x.num = a.num;

      elseif isa(a, 'integer')

        x.num = zeros(exfloat.size64,1, 'uint64');
        exflib_float_ltoef(x.num, a, exfloat.precision64);

      elseif isa(a, 'char')

        x.num = zeros(exfloat.size64,1, 'uint64');
        exflib_float_etoef(a, x.num, exfloat.precision64);

      elseif (isa(a,'double') || isa(a,'float')) ...
            && (a-fix(a) == 0)  % essentially integer

        x.num = zeros(exfloat.size64,1, 'uint64');
        exflib_float_ltoef(x.num, int64(a), exfloat.precision64);

      elseif row~=1 | col~=1

        error('conversion from array not yet implemented');
        return;

      else

        error('Implicit convertion from %d is prohibited.', a);
        return;

      end % if

    end % function x=exfloat(a)

  end % methods

  methods

    function cast( x, d )
    
      % explicit conversion from double to exfloat
      % usage:
      %    f = exfloat();
      %    f.cast(0.1);
      
      if isa(d, 'double')
        exflib_float_dtoef(d, x.num, exfloat.precision64);
      else
        error('exflib_cast: conversion from unknown type');
      end
      
    end % function cast

  end % methods

  methods (Static)

    %------------------------------
    % auxiliary routines
    %------------------------------
    function p = get_precision()
      p = exfloat.precision;
    end

    function p = get_req_precision()
      p = exfloat.precision10;
    end	  

    function p = get_exfloat_precision64()
      p = exfloat.precision64;
    end

    function p = get_exfloat_byte()
      p = exfloat.byte;
    end

  end % methods(Static)

  methods (Static, Access=private)

    function [ output_prec, output_fmt ] = parse_fmt ( fmt )

      % default
      output_prec = exfloat.default_output_precision;
      output_fmt  = exfloat.default_output_format; % default

      if ~isa(fmt, 'char')
        %warning('fmt should be integer : %s', fmt) 
       return
      end

      if ( fmt(1)=='%' && (fmt(2)=='f' || fmt(2)=='F') )
        output_fmt = 'f'; % fixed format
        return
      end

      if ( fmt(1)=='%' && (fmt(2)=='e' || fmt(2)=='E') )
        output_fmt = 'e'; % scientific format
        return
      end

      if ( fmt(1)=='%' && (fmt(2)=='g' || fmt(2)=='G') )
        output_fmt = 'g'; % generic format';
        return
      end
	
      if ~( fmt(1)=='%' && fmt(2)=='.' )
        %warning('illeagal format : %s', fmt)
        return
      end

      output_prec = 0;
      for n=3:length(fmt)

        d = fmt(n) - '0';

        if (d < 0) || (9 < d)
          break
        end
    
        output_prec = output_prec * 10;
        output_prec = output_prec + d;
      end

      if fmt(n) == 'e' || fmt(n) == 'E'
        output_fmt = 'e'; % scientific format
      elseif fmt(n) == 'f' || fmt(n) == 'F'
        output_fmt = 'f'; % fixed format
      elseif fmt(n) == 'g' || fmt(n) == 'G'
        output_fmt = 'g'; % generic format';
      else
        %warning('unknown format : %s', fmt)
      end

    end % [output_prec,output_fmt]=parse_fmt(fmt)

  end % methods (Access=private)

end % classdef

%----------------------------------------------------------------------
% End of file
%----------------------------------------------------------------------
