/* -*- C -*- */
/* Time-stamp: "2015-11-08 16:52:35 user" */
/*----------------------------------------------------------------------*/
/*  Copyright (C) 2015, FUJIWARA Hiroshi                                */
/*    <fujiwara@acs.i.kyoto-u.ac.jp>                                    */
/*----------------------------------------------------------------------*/
#include "mex.h"
#include "exfloat.h"

#ifdef _OPENMP
#include <omp.h>
#endif

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  unsigned long long *const a = (unsigned long long*)mxGetData( prhs[0] );  
  const size_t N    = mxGetScalar( prhs[1] );  /* size */
  const size_t M    = mxGetScalar( prhs[2] );  /* number of RHS */
  const size_t pr64 = mxGetScalar( prhs[3] );  /* precision64 */
  unsigned long long *const w = (unsigned long long*)mxGetData( prhs[4] );
  
  const size_t s64 = 1 + pr64;
  unsigned long long *const w0 = w;
  unsigned long long *const w1 = w + (1+pr64);
  
#if 0
  printf("a = %d x %d\n", N, N);
  printf("b = %d x %d\n", N, M);
  printf("prec = %d, size = %d\n", pr64, s64);
#endif
  
  const size_t NM = N + M;
  int i, j, k, l, p;

#if 0
  /******************************************************/
  /* debug                                              */
  /******************************************************/
  for(i = 0; i < N; i++){
      for(j = 0; j < N; j++){
          printf("a(%d,%d)=", i,j);
          for(p = 0; p <= pr64; p++){
	    printf("%016lx:", a[(i*NM+j)*s64+p]);
          }
          printf("\n");
      }
  }
  
  for(i = 0; i < N; i++){
      for(j = 0; j < M; j++){
          printf("b(%d,%d)=", i,j);
          for(p = 0; p <= pr64; p++){
	    printf("%016lx:", a[(i*NM+N+j)*s64+p]);
          }
          printf("\n");
      }
  }
  /******************************************************/
#endif

  /* a[i][j] corresponds to a[ (i*NM+j)*s64 ] */

  /* forward elimination */
  for(k = 0; k < N-1; k++){
      
      /* pivoting */
      int max_idx = k;
      for(p = 0; p < s64; p++)
	  w0[p] = a[(k*NM+k)*s64 + p];
      w0[0] &= 0x7fffffffffffffffULL;  /* abs( a[k][k] ) */

      /* finding pivot */
      for(l = k+1; l < N; l++){
	
	for(p = 0; p < s64; p++)
	  w1[p] = a[(l*NM+k)*s64 + p];
	w1[0] &= 0x7fffffffffffffffULL; /* abs( a[l][k] ) */

	if( exflib_float_lt(w0, w1, pr64) ){
	    max_idx = l;
	    for(p = 0; p < s64; p++)
	      w0[p] = w1[p];
	}
      }

      /* exchanging pivot */
      if( k != max_idx ){
	for(j = k; j < NM; j++){
	  for(p = 0; p < s64; p++)
	    w0[p] = a[(k*NM+j)*s64+p];
	  for(p = 0; p < s64; p++)
	    a[(k*NM+j)*s64+p] = a[(max_idx*NM+j)*s64+p];
	  for(p = 0; p < s64; p++)
	    a[(max_idx*NM+j)*s64+p] = w0[p];
	}
      }

      for(i = k+1; i < N; i++){

	/* w0 = a[i][k] / a[k][k]; */
	exflib_float_div(w0, a+(i*NM+k)*s64, a+(k*NM+k)*s64, pr64);

        #pragma omp parallel for
        for(j = k+1; j < NM; j++){
	  /* a[i][j] -= w0 * a[k][j]; */
	  exflib_float_mul(w1, w0, a+(k*NM+j)*s64, pr64);
	  exflib_float_sub(a+(i*NM+j)*s64, a+(i*NM+j)*s64, w1, pr64);
	}

        /* a[i][k] = 0; */
	for(p = 0; p < s64; p++)
	  a[(i*NM+k)*s64 + p] = 0;
      }
  }
  
  /* backward substitution */
  for(k = N-1; k >= 0; k--){
    for(j = 0; j < M; j++){
      
      for(l = k+1; l < N; l++){
	/* a[k][N+j] -= a[k][l] * a[l][N+j]; */
	exflib_float_mul(w0, a + (k*NM+l)*s64, a + (l*NM+N+j)*s64, pr64);
	exflib_float_sub(a + (k*NM+N+j)*s64, a + (k*NM+N+j)*s64, w0, pr64);
      }

      /* a[k][N+j] /= a[k][k]; */
      exflib_float_div(a + (k*NM+N+j)*s64, a + (k*NM+N+j)*s64, a + (k*NM+k)*s64, pr64);
    }
  }
  
  /* exflib_ulong roundinfo = exflib_float_add( c, a, b, N ); */
  /* exflib_float_round_near(c, roundinfo, N); */

  return;
}

/* End of file */
