/* -*- C -*- */
/* Time-stamp: "2018-08-16 19:59:17 fujiwara" */
/*----------------------------------------------------------------------*/
/*  Copyright (C) 2018, FUJIWARA Hiroshi                                */
/*    <fujiwara@acs.i.kyoto-u.ac.jp>                                    */
/*----------------------------------------------------------------------*/
#include "mex.h"
#include "exfloat.h"
#include <assert.h>

/*----------------------------------------------------------------------*/
/* The following routine is copy from JAMA (A Java Matrix Package)      */
/* https://math.nist.gov/javanumerics/jama/                             */
/*----------------------------------------------------------------------*/

/*
  void QR( exflib_uQword *const &A, const int row, const int col, const int pr64,
           exflib_uQword *const &Q, exflib_uQword *const &R,
           const bool wantq = true, const bool wantr = true );

  JAMA  : A=QR, Q: m-by-n (m >= n), orthogonal, R: n-by-n, triangular
  MATLAB: A=QR, Q: m-by-m, unitary, R: m-by-n, upper triangular
*/
  
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    unsigned long long *const A   = (unsigned long long*)mxGetData( prhs[0] );  
    const size_t row              = mxGetScalar( prhs[1] );
    const size_t col              = mxGetScalar( prhs[2] );
    const size_t pr64             = mxGetScalar( prhs[3] );              /* precision64 */
    unsigned long long *const Q   = (unsigned long long*)mxGetData( prhs[4] );
    unsigned long long *const R   = (unsigned long long*)mxGetData( prhs[5] );

    const int wantq = mxGetScalar( prhs[6] );
    const int wantr = mxGetScalar( prhs[7] );

    const int min_row_col   = (row < col)? row: col;       /* min(row,col); */

    const size_t s64   = 1 + pr64;
    const int bytes    = s64 * sizeof(exflib_uQword);
    const size_t guard = 1;
    const size_t s64g  = 1 + pr64 + guard;

#ifdef __cplusplus
    exflib_uQword *const work_sqrt = new exflib_uQword [ s64g*4 ];
    exflib_uQword *const tmp = new exflib_uQword [ s64*(3+col) ];
#else
    const int bytesg = s64g * sizeof(exflib_uQword);
    exflib_uQword *const work_sqrt = (exflib_uQword*)malloc( bytesg*4 );
    exflib_uQword *const tmp = (exflib_uQword*)malloc( bytes*(3+col) );
#endif
    exflib_uQword *const s     = tmp + s64*1;
    exflib_uQword *const nrm   = tmp + s64*2;
    exflib_uQword *const Rdiag = tmp + s64*3;

    int i = 0, j = 0, k = 0;
    size_t p;

    /* assert( row >= col ); */

    /* Main loop. */
    for(k = 0; k < col; k++) {
	/* Compute 2-norm of k-th column without under/overflow. */
	for(p = 0; p < s64; p++)
	    nrm[p] = 0;
	for(i = k; i < row; i++) {
	    exflib_float_mul(tmp, A+(i*col+k)*s64, A+(i*col+k)*s64, pr64);
	    exflib_float_add(nrm, nrm, tmp, pr64);
	}
	exflib_float_sqrt(nrm, nrm, pr64, work_sqrt);

	if( (nrm[0] & EXFLIB_MASK_EXP) != 0 ){
            /* Form k-th Householder vector. */
	    /* original requirement: col < row, since k == col-1 */
            if( ((A[(k*col+k)*s64] & EXFLIB_MASK_SGN) != 0) && (A[(k*col+k)*s64] != EXFLIB_MASK_SGN) ) {
		nrm[0] ^= EXFLIB_MASK_SGN;
            }
            for(i = k; i < row; i++) {
		exflib_float_div(A+(i*col+k)*s64, A+(i*col+k)*s64, nrm, pr64);
            }
	    tmp[0] = EXFLIB_BIAS;
	    for(p = 1; p < s64; p++)
		tmp[p] = 0;

	    /* original requirement: col < row, since k == col-1 */
	    exflib_float_add(A+(k*col+k)*s64, A+(k*col+k)*s64, tmp, pr64);

            /* Apply transformation to remaining columns. */
            for(j = k+1; j < col; j++) {
		for(p = 0; p < s64; p++)
		    s[p] = 0; 
		for(i = k; i < row; i++) {
		    exflib_float_mul(tmp, A+(i*col+k)*s64, A+(i*col+j)*s64, pr64);
		    exflib_float_add(s, s, tmp, pr64);
		}
		/* original requirement: col < row, since k == col-1 */
		exflib_float_div(s, s, A+(k*col+k)*s64, pr64);
		s[0] ^= EXFLIB_MASK_SGN;
		for(i = k; i < row; i++) {
		    exflib_float_mul(tmp, s, A+(i*col+k)*s64, pr64);
		    exflib_float_add(A+(i*col+j)*s64, A+(i*col+j)*s64, tmp, pr64);
		}
            }
	}
	Rdiag[k*s64] = nrm[0] ^ EXFLIB_MASK_SGN;
	for(p = 1; p < s64; p++)
	    Rdiag[k*s64+p] = nrm[p];
    }

    /*--------------------------------------------------*/
    /* getQ : Q[ROW][COL]                               */
    /*--------------------------------------------------*/
    if( wantq ){
	for(k = min_row_col-1; k >= 0; k--) {
	    for(i = 0; i < row; i++) {
		for(p = 0; p < s64; p++)
		    Q[(i*col+k)*s64+p] = 0;
	    }
	
	    /* original requirement: col < row, since k == col-1 */
	    Q[(k*col+k)*s64] = EXFLIB_BIAS;
	    for(p = 1; p < s64; p++)
		Q[(k*col+k)*s64+p] = 0;

	    for(j = k; j < col; j++) {

		/* original requirement: col < row, since k == col-1 */
		if( (A[(k*col+k)*s64] & EXFLIB_MASK_EXP) != 0) {

		    for(p = 0; p < s64; p++)
			s[p] = 0;

		    for(i = k; i < row; i++) {
			exflib_float_mul(tmp, A+(i*col+k)*s64, Q+(i*col+j)*s64, pr64);
			exflib_float_add(s, s, tmp, pr64);
		    }
		    /* original requirement: col < row, since k == col-1 */
		    exflib_float_div(s, s, A+(k*col+k)*s64, pr64);
		    s[0] ^= EXFLIB_MASK_SGN;
		    for(i = k; i < row; i++) {
			exflib_float_mul(tmp, s, A+(i*col+k)*s64, pr64);
			exflib_float_add(Q+(i*col+j)*s64, Q+(i*col+j)*s64, tmp, pr64);
		    }
		}
	    }
	}
    }

    /*--------------------------------------------------*/
    /* getR : R[COL][COL]                               */
    /*--------------------------------------------------*/
    if( wantr ){
	for(i = 0; i < min_row_col; i++) {
	    for(j = 0; j < col; j++) {
		if(i < j) {
		    /* original requirement: col < row, since i == col-1 */
		    for(p = 0; p < s64; p++)
			R[(i*col+j)*s64+p] = A[(i*col+j)*s64+p];
		} else if(i == j) {
		    for(p = 0; p < s64; p++)
			R[(i*col+j)*s64+p] = Rdiag[i*s64+p];
		} else {
		    for(p = 0; p < s64; p++)
			R[(i*col+j)*s64+p] = 0;
		}
	    }
	}
    } else { /* wantr == false */
	for(i = 0; i < min_row_col; i++)
	    for(p = 0; p < s64; p++)
		A[(i*col+i)*s64+p] = Rdiag[i*s64+p];
    }

#ifdef __cplusplus
    delete [] work_sqrt;
    delete [] tmp;
#else
    free( work_sqrt );
    free( tmp );
#endif

    return;
}

/* End of file */
