/* -*- C -*- */
/* Time-stamp: "2018-08-15 09:34:10 fujiwara" */
/*----------------------------------------------------------------------*/
/*  Copyright (C) 2018, FUJIWARA Hiroshi                                */
/*    <fujiwara@acs.i.kyoto-u.ac.jp>                                    */
/*----------------------------------------------------------------------*/
#include "mex.h"
#include "exfloat.h"

/*----------------------------------------------------------------------*/
/* The following routine is copy from JAMA (A Java Matrix Package)      */
/* https://math.nist.gov/javanumerics/jama/                             */
/*----------------------------------------------------------------------*/

static void exflib_mex_hypot( exflib_uQword *const s,
			      const exflib_uQword *const a,
			      const exflib_uQword *const b, const size_t pr64 )
{
    const size_t guard = 1;
    const size_t pr64g = pr64 + guard;
#ifdef __cplusplus
    exflib_uQword *work = new exflib_uQword [ (1+pr64g)*4 ];
#else
    const int bytes = (1+pr64g) * sizeof(exflib_uQword);
    exflib_uQword *const work = (exflib_uQword*)malloc( bytes*4 );
#endif

    exflib_float_mul(s, a, a, pr64);
    exflib_float_mul(work, b, b, pr64);
    exflib_float_add(s, s, work, pr64);
    exflib_float_sqrt(s, s, pr64, work);

#ifdef __cplusplus
    delete [] work;
#else
    free( work );
#endif
    return;
}

/*
  void SVD( exflib_uQword *const &A, const int row, const int col, const int pr64,
            exflib_uQword *const &U, exflib_uQword *const &S, exflib_uQword *const &V,
            const bool wantu = true, const bool wantv = true );
*/
  
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    unsigned long long *const A   = (unsigned long long*)mxGetData( prhs[0] );  
    const size_t row              = mxGetScalar( prhs[1] );
    const size_t col              = mxGetScalar( prhs[2] );
    const size_t pr64             = mxGetScalar( prhs[3] );              /* precision64 */
    unsigned long long *const U   = (unsigned long long*)mxGetData( prhs[4] );
    unsigned long long *const S   = (unsigned long long*)mxGetData( prhs[5] );
    unsigned long long *const V   = (unsigned long long*)mxGetData( prhs[6] );

    const int minm = (row+1 < col)? (row+1): col; /* min(row+1,col); */
    const int nu   = (row < col)? row: col;       /* min(row,col); */

    const int wantu = mxGetScalar( prhs[7] );
    const int wantv = mxGetScalar( prhs[8] );
  
    const size_t s64   = 1 + pr64;
    const int bytes    = s64 * sizeof(exflib_uQword);
    const size_t guard = 1;
    const size_t s64g  = 1 + pr64 + guard;
    const int bytesg   = s64g * sizeof(exflib_uQword);

#ifdef __cplusplus
    exflib_uQword *const e    = new exflib_uQword [col*s64];
    exflib_uQword *const work = new exflib_uQword [row*s64];
    exflib_uQword *const work_sqrt = new exflib_uQword [ s64g*4 ];
    exflib_uQword *const tmp  = new exflib_uQword [ s64*6 ];
#else
    exflib_uQword *const e    = (exflib_uQword*)malloc( col*bytes );
    exflib_uQword *const work = (exflib_uQword*)malloc( row*bytes );
    exflib_uQword *const work_sqrt = (exflib_uQword*)malloc( bytesg*4 );
    exflib_uQword *const tmp  = (exflib_uQword*)malloc( bytes*6 );
#endif
    exflib_uQword *const tmp2 = tmp + s64;
    exflib_uQword *const eps  = tmp + s64*2;
    exflib_uQword *const t    = tmp + s64*3;
    exflib_uQword *const cs   = tmp + s64*4;
    exflib_uQword *const sn   = tmp + s64*5;

    int i=0, j=0, k=0;
    size_t f;
    int p, pp, iter;

    /* Reduce A to bidiagonal form, storing the diagonal elements */
    /* in s and the super-diagonal elements in e. */

    const int nct = (row-1 < col)? (row-1): col; /* min(row-1,col) */

    const int nrtm = (col-2 < row)? (col-2): row; /* min(col-2,row) */
    const int nrt  = (0 > nrtm)? 0: nrtm; /* max(0,min(col-2,row)); */

    const int maxitr = (nct > nrt)? nct: nrt; /* max(nct,nrt) */

    for(k = 0; k < maxitr; k++) {
	if(k < nct) {
            /* Compute the transformation for the k-th column and */
            /* place the k-th diagonal in s[k]. */
            /* Compute 2-norm of k-th column without under/overflow. */

	    for(f = 0; f < s64; f++)
		tmp[f] = 0;
            for(i = k; i < row; i++) {
		exflib_float_mul(tmp2, A+(i*col+k)*s64, A+(i*col+k)*s64, pr64);
		exflib_float_add(tmp, tmp, tmp2, pr64);
            }
	    exflib_float_sqrt(S+k*s64, tmp, pr64, work_sqrt);

            if( (S[k*s64] & EXFLIB_MASK_EXP) != 0 ) {
		if((A[(k*col+k)*s64] & EXFLIB_MASK_SGN) != 0) {
		    S[k*s64] ^= EXFLIB_MASK_SGN;
		}
		for(i = k; i < row; i++) {
		    exflib_float_div(A+(i*col+k)*s64, A+(i*col+k)*s64, S+k*s64, pr64);
		}
		tmp[0] = EXFLIB_BIAS;
		for(f = 1; f < s64; f++)
		    tmp[f] = 0;

		exflib_float_add(A+(k*col+k)*s64, A+(k*col+k)*s64, tmp, pr64);
            }
            S[k*s64] ^= EXFLIB_MASK_SGN;
	}
	for(j = k+1; j < col; j++) {
            if((k < nct) && ( (S[k*s64] & EXFLIB_MASK_EXP) != 0) ) {

		/* Apply the transformation. */

		t[0] = 0;
		for(f = 1; f < s64; f++)
		    t[f] = 0;
		for(i = k; i < row; i++) {
		    exflib_float_mul(tmp, A+(i*col+k)*s64, A+(i*col+j)*s64, pr64);
		    exflib_float_add(t, t, tmp, pr64);
		}
		exflib_float_div(t, t, A+(k*col+k)*s64, pr64);
		t[0] ^= EXFLIB_MASK_SGN;
		for(i = k; i < row; i++) {
		    exflib_float_mul(tmp, t, A+(i*col+k)*s64, pr64);
		    exflib_float_add(A+(i*col+j)*s64, A+(i*col+j)*s64, tmp, pr64);
		}
            }

            /* Place the k-th row of A into e for the */
            /* subsequent calculation of the row transformation. */

	    for(f = 0; f < s64; f++)
		e[j*s64+f] = A[(k*col+j)*s64+f];
	}
	if(wantu & (k < nct)) {

            /* Place the transformation in U for subsequent back */
            /* multiplication. */

            for(i = k; i < row; i++) {
		for(f = 0; f < s64; f++)
		    U[(i*nu+k)*s64+f] = A[(i*col+k)*s64+f];
            }
	}
	if(k < nrt) {

            /* Compute the k-th row transformation and place the */
            /* k-th super-diagonal in e[k]. */
            /* Compute 2-norm without under/overflow. */

	    for(f = 0; f < s64; f++)
		tmp[f] = 0;
	    for(i = k+1; i < col; i++){
		exflib_float_mul(tmp2, e+i*s64, e+i*s64, pr64);
		exflib_float_add(tmp, tmp, tmp2, pr64);
	    }
	    exflib_float_sqrt(e+k*s64, tmp, pr64, work_sqrt);

            if( (e[k*s64] & EXFLIB_MASK_EXP) != 0) {
		if( (e[(k+1)*s64] & EXFLIB_MASK_SGN) != 0) {
		    e[k*s64] ^= EXFLIB_MASK_SGN;
		}
		for(i = k+1; i < col; i++) {
		    exflib_float_div(e+(i)*s64, e+(i)*s64, e+k*s64, pr64);
		}
		tmp[0] = EXFLIB_BIAS;
		for(f = 1; f < s64; f++)
		    tmp[f] = 0;

		exflib_float_add(e+(k+1)*s64, e+(k+1)*s64, tmp, pr64);
            }
            e[k*s64] ^= EXFLIB_MASK_SGN;

            if((k+1 < row) && ( (e[k*s64] & EXFLIB_MASK_EXP) != 0) ) {

		/* Apply the transformation. */

		for(i = k+1; i < row; i++) {
		    for(f = 0; f < s64; f++)
			work[(i)*s64+f] = 0;
		}
		for(j = k+1; j < col; j++) {
		    for(i = k+1; i < row; i++) {
			exflib_float_mul(tmp, e+j*s64, A+(i*col+j)*s64, pr64);
			exflib_float_add(work+(i)*s64, work+(i)*s64, tmp, pr64);
		    }
		}
		for(j = k+1; j < col; j++) {
		    exflib_float_div(t, e+j*s64, e+(k+1)*s64, pr64);
		    t[0] ^= EXFLIB_MASK_SGN;
		    for(i = k+1; i < row; i++) {
			exflib_float_mul(tmp, t, work+(i)*s64, pr64);
			exflib_float_add(A+(i*col+j)*s64, A+(i*col+j)*s64, tmp, pr64);
		    }
		}
            }
            if(wantv) {

		/* Place the transformation in V for subsequent */
		/* back multiplication. */

		for(i = k+1; i < col; i++) {
		    for(f = 0; f < s64; f++)
			V[(i*col+k)*s64+f] = e[(i)*s64+f];
		}
            }
	}
    }
    /* Set up the final bidiagonal matrix or order p. */
    p = (col < row+1)? col: (row+1); /* min(col,row+1); */
    if(nct < col) {
	for(f = 0; f < s64; f++)
	    S[(nct)*s64+f] = A[(nct*col+nct)*s64+f];
    }
    if(row < p) {
	for(f = 0; f < s64; f++)
	    S[(p-1)*s64+f] = 0;
    }
    if(nrt+1 < p) {
	for(f = 0; f < s64; f++)
	    e[(nrt)*s64+f] = A[(nrt*col+p-1)*s64+f];
    }
    for(f = 0; f < s64; f++)
	e[(p-1)*s64+f] = 0;

    /* If required, generate U. */

    if(wantu) {
	for(j = nct; j < nu; j++) {
            for(i = 0; i < row; i++) {
		for(f = 0; f < s64; f++)
		    U[(i*nu+j)*s64+f] = 0;
            }

            U[(j*nu+j)*s64] = EXFLIB_BIAS;
	    for(f = 1; f < s64; f++)
		U[(j*nu+j)*s64+f] = 0;
	    
	}
	for(k = nct-1; k >= 0; k--) {
            if( (S[k*s64] & EXFLIB_MASK_EXP) != 0 ){
		for(j = k+1; j < nu; j++) {

		    for(f = 0; f < s64; f++)
			t[f] = 0;
		    for(i = k; i < row; i++) {
			exflib_float_mul(tmp, U+(i*nu+k)*s64, U+(i*nu+j)*s64, pr64);
			exflib_float_add(t, t, tmp, pr64);
		    }
		    exflib_float_div(t, t, U+(k*nu+k)*s64, pr64);
		    t[0] ^= EXFLIB_MASK_SGN;
		    for(i = k; i < row; i++) {
			exflib_float_mul(tmp, t, U+(i*nu+k)*s64, pr64);
			exflib_float_add(U+(i*nu+j)*s64, U+(i*nu+j)*s64, tmp, pr64);
		    }
		}
		for(i = k; i < row; i++ ) {
		    U[(i*nu+k)*s64] ^= EXFLIB_MASK_SGN;
		}
		tmp[0] = EXFLIB_BIAS;
		for(f = 1; f < s64; f++)
		    tmp[f] = 0;
		exflib_float_add(U+(k*nu+k)*s64, U+(k*nu+k)*s64, tmp, pr64);
		for(i = 0; i < k-1; i++) {
		    for(f = 0; f < s64; f++)
			U[(i*nu+k)*s64+f] = 0;
		}
            } else {
		for(i = 0; i < row; i++) {
		    for(f = 0; f < s64; f++)
			U[(i*nu+k)*s64+f] = 0;
		}
		U[(k*nu+k)*s64] = EXFLIB_BIAS;
		for(f = 1; f < s64; f++)
		    U[(k*nu+k)*s64+f] = 0;
            }
	}
    }

    /* If required, generate V. */
    
    if(wantv) {
	for(k = col-1; k >= 0; k--) {
            if((k < nrt) && ((e[k*s64] & EXFLIB_MASK_EXP) != 0) ) {
		for(j = k+1; j < nu; j++) {
		    for(f = 0; f < s64; f++)
			t[f] = 0;
		    for(i = k+1; i < col; i++) {
			exflib_float_mul(tmp, V+(i*col+k)*s64, V+(i*col+j)*s64, pr64);

			exflib_float_add(t, t, tmp, pr64);
		    }
		    exflib_float_div(t, t, V+((k+1)*col+k)*s64, pr64);
		    t[0] ^= EXFLIB_MASK_SGN;
		    for(i = k+1; i < col; i++) {
			exflib_float_mul(tmp, t, V+(i*col+k)*s64, pr64);
			exflib_float_add(V+(i*col+j)*s64, V+(i*col+j)*s64, tmp, pr64);
		    }
		}
            }
	    for(i = 0; i < col; i++) {
		for(f = 0; f < s64; f++)
		    V[(i*col+k)*s64+f] = 0;
	    }
	    V[(k*col+k)*s64] = EXFLIB_BIAS;
	    for(f = 1; f < s64; f++)
		V[(k*col+k)*s64+f] = 0;
	}
    }

    /* Main iteration loop for the singular values. */

    pp = p-1;
    iter = 0;
    /*eps[0] = EXFLIB_BIAS - 52; */
    eps[0] = EXFLIB_BIAS - 64*pr64;
    for(f = 1; f < s64; f++)
	eps[f] = 0;

    while(p > 0) {
	int k=0;
	int kase=0;

	/* Here is where a test for too many iterations would go. */
	/*
	   This section of the program inspects for
	   negligible elements in the s and e arrays.  On
	   completion the variables kase and k are set as follows.

	   kase = 1     if s(p) and e[k-1] are negligible and k<p
	   kase = 2     if sk is negligible and k<p
	   kase = 3     if e[k-1] is negligible, k<p, and
	                sk, ..., s(p) are not negligible (qr step).
	   kase = 4     if e(p-1) is negligible (convergence).
	*/

	for(k = p-2; k >= -1; k--) {
            if(k == -1) {
		break;
            }

	    tmp[0] = S[k*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp[f] = S[k*s64+f];
	    tmp2[0] = S[(k+1)*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp2[f] = S[(k+1)*s64+f];
	    exflib_float_add(tmp2, tmp2, tmp, pr64);

	    exflib_float_mul(tmp2, tmp2, eps, pr64);

	    tmp[0] = e[k*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp[f] = e[k*s64+f];
	    
            if(!exflib_float_lt(tmp2, tmp, pr64)){
		e[k*s64] = 0;
		break;
            }
	}
	if(k == p-2) {
            kase = 4;
	} else {
            int ks;
            for(ks = p-1; ks >= k; ks--) {
		if(ks == k) {
		    break;
		}

		if(ks != p){
		    tmp[0] = e[ks*s64] & EXFLIB_MASK_EXP;
		    for(f = 1; f < s64; f++)
			tmp[f] = e[ks*s64+f];
		}else{
		    for(f = 0; f < s64; f++)
			tmp[f] = 0;
		}
		if(ks != k+1){
		    tmp2[0] = e[(ks-1)*s64] & EXFLIB_MASK_EXP;
		    for(f = 1; f < s64; f++)
			tmp2[f] = e[(ks-1)*s64+f];
		}else{
		    for(f = 0; f < s64; f++)
			tmp2[f] = 0;
		}

		exflib_float_add(t, tmp, tmp2, pr64);

		tmp[0] = S[ks*s64] & EXFLIB_MASK_EXP;
		for(f = 1; f < s64; f++)
		    tmp[f] = S[ks*s64+f];
		exflib_float_mul(tmp2, eps, t, pr64);

		if(!exflib_float_lt(tmp2, tmp, pr64)){
		    S[ks] = 0;
		    break;
		}
            }
            if(ks == k) {
		kase = 3;
            } else if(ks == p-1) {
		kase = 1;
            } else {
		kase = 2;
		k = ks;
            }
	}
	k++;

	/* Perform the task indicated by kase. */

	switch (kase) {

            /* Deflate negligible s(p). */

	case 1: {
	    for(f = 0; f < s64; f++)
		tmp2[f] = e[(p-2)*s64+f];
	    for(f = 0; f < s64; f++)
		e[(p-2)*s64+f] = 0;
	    for(j = p-2; j >= k; j--) {
		exflib_mex_hypot(t, S+j*s64, tmp2, pr64);
		exflib_float_div(cs, S+j*s64, t, pr64);
		exflib_float_div(sn, tmp2, t, pr64);
		for(f = 0; f < s64; f++)
		    S[j*s64+f] = t[f];
		if(j != k) {
		    exflib_float_mul(tmp2, sn, e+(j-1)*s64, pr64);
		    tmp2[0] ^= EXFLIB_MASK_SGN;
		    exflib_float_mul(e+(j-1)*s64, cs, e+(j-1)*s64, pr64);
		}
		if(wantv) {
		    for(i = 0; i < col; i++) {

			exflib_float_mul(t, cs, V+(i*col+j)*s64, pr64);
			exflib_float_mul(tmp, sn, V+(i*col+p-1)*s64, pr64);
			exflib_float_add(t, t, tmp, pr64);

			exflib_float_mul(V+(i*col+p-1)*s64, cs, V+(i*col+p-1)*s64, pr64);
			exflib_float_mul(tmp, sn, V+(i*col+j)*s64, pr64);
			exflib_float_sub(V+(i*col+p-1)*s64, V+(i*col+p-1)*s64, tmp, pr64);

			for(f = 0; f < s64; f++)
			    V[(i*col+j)*s64+f] = t[f];
		    }
		}
	    }
	}
            break;

            /* Split at negligible sk. */

	case 2: {
	    for(f = 0; f < s64; f++)
		tmp2[f] = e[(k-1)*s64+f];
	    for(f = 0; f < s64; f++)
		e[(k-1)*s64+f] = 0;
	    for(j = k; j < p; j++) {
		exflib_mex_hypot(t, S+j*s64, tmp2, pr64);
		exflib_float_div(cs, S+j*s64, t, pr64);
		exflib_float_div(sn, tmp2, t, pr64);
		for(f = 0; f < s64; f++)
		    S[j*s64+f] = t[0];
		/* ff = -sn*e[j]; */ /* not used */
		exflib_float_mul(e+j*s64, cs, e+j*s64, pr64);
		if(wantu) {
		    for(i = 0; i < row; i++) {
                        exflib_float_mul(t, cs, U+(i*nu+j)*s64, pr64);
			exflib_float_mul(tmp, sn, U+(i*nu+k-1)*s64, pr64);
			exflib_float_add(t, t, tmp, pr64);

			exflib_float_mul(U+(i*nu+k-1)*s64, cs, U+(i*nu+k-1)*s64, pr64);
			exflib_float_mul(tmp, sn, U+(i*nu+j)*s64, pr64);
			exflib_float_sub(U+(i*nu+k-1)*s64, U+(i*nu+k-1)*s64, tmp, pr64);

			for(f = 0; f < s64; f++)
			    U[(i*nu+j)*s64+f] = t[f];
		    }
		}
	    }
	}
            break;

            /* Perform one qr step. */

	case 3: {

#ifdef __cplusplus
	    exflib_uQword *const work3 = new exflib_uQword [ s64*12 ];
#else
	    exflib_uQword *const work3 = (exflib_uQword*)malloc( bytes*12 );
#endif
	    exflib_uQword *const recip = work3 + s64*1;
	    exflib_uQword *const sp    = work3 + s64*2;
	    exflib_uQword *const spm1  = work3 + s64*3;
	    exflib_uQword *const epm1  = work3 + s64*4;
	    exflib_uQword *const sk    = work3 + s64*5;
	    exflib_uQword *const ek    = work3 + s64*6;
	    exflib_uQword *const b     = work3 + s64*7;
	    exflib_uQword *const c     = work3 + s64*8;
	    exflib_uQword *const shift = work3 + s64*9;
	    exflib_uQword *const h     = work3 + s64*10;
	    exflib_uQword *const g     = work3 + s64*11;
	    
	    /* Calculate the shift. */
   
	    tmp[0] = S[(p-1)*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp[f] = S[(p-1)*s64+f];

	    tmp2[0] = S[(p-2)*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp2[f] = S[(p-2)*s64+f];
	    if(exflib_float_lt(tmp, tmp2, pr64)){
		for(f = 0; f < s64; f++)
		    tmp[f] = tmp2[f];
	    }
	    
	    tmp2[0] = e[(p-2)*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp2[f] = e[(p-2)*s64+f];
	    if(exflib_float_lt(tmp, tmp2, pr64)){
		for(f = 0; f < s64; f++)
		    tmp[f] = tmp2[f];
	    }

	    tmp2[0] = S[k*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp2[f] = S[k*s64+f];
	    if(exflib_float_lt(tmp, tmp2, pr64)){
		for(f = 0; f < s64; f++)
		    tmp[f] = tmp2[f];
	    }
	    
	    tmp2[0] = e[k*s64] & EXFLIB_MASK_EXP;
	    for(f = 1; f < s64; f++)
		tmp2[f] = e[k*s64+f];
	    if(exflib_float_lt(tmp, tmp2, pr64)){
		for(f = 0; f < s64; f++)
		    tmp[f] = tmp2[f];
	    }
	    
	    tmp2[0] = EXFLIB_BIAS;
	    for(f = 1; f < s64; f++)
		tmp2[f] = 0;
	    exflib_float_div(recip, tmp2, tmp, pr64);

	    exflib_float_mul(sp, S+(p-1)*s64, recip, pr64);
	    exflib_float_mul(spm1, S+(p-2)*s64, recip, pr64);
	    exflib_float_mul(epm1, e+(p-2)*s64, recip, pr64);
	    exflib_float_mul(sk, S+k*s64, recip, pr64);
	    exflib_float_mul(ek, e+k*s64, recip, pr64);

	    exflib_float_add(tmp, spm1, sp, pr64);
	    exflib_float_sub(tmp2, spm1, sp, pr64);
	    exflib_float_mul(tmp, tmp, tmp2, pr64);
	    exflib_float_mul(b, epm1, epm1, pr64);
	    exflib_float_add(b, b, tmp, pr64);

	    exflib_float_mul(c, sp, epm1, pr64);
	    exflib_float_mul(c, c, c, pr64);

	    for(f = 0; f < s64; f++)
		shift[f] = tmp[0] = 0;

	    if( ((b[0] & EXFLIB_MASK_EXP) != 0) || ((c[0] & EXFLIB_MASK_EXP) != 0) ){

		exflib_float_mul(tmp, b, b, pr64);
		exflib_float_add(tmp, tmp, c, pr64);
		exflib_float_sqrt(shift, tmp, pr64, work_sqrt);

		if( ((b[0] & EXFLIB_MASK_SGN) != 0) && (b[0] != EXFLIB_MASK_SGN) ){
		    shift[0] ^= EXFLIB_MASK_SGN;
		}
		exflib_float_add(tmp, b, shift, pr64);
		exflib_float_div(shift, c, tmp, pr64);
	    }
	    
	    exflib_float_add(tmp, sk, sp, pr64);
	    exflib_float_sub(tmp2, sk, sp, pr64);
	    exflib_float_mul(tmp, tmp, tmp2, pr64); 
	    exflib_float_add(h, tmp, shift, pr64);
	    exflib_float_mul(g, sk, ek, pr64);
   
	    /* Chase zeros. */
   
	    for(j = k; j < p-1; j++) {
		exflib_mex_hypot(t, h, g, pr64);
		exflib_float_div(cs, h, t, pr64);
		exflib_float_div(sn, g, t, pr64);
		if(j != k) {
		    for(f = 0; f < s64; f++)
			e[(j-1)*s64+f] = t[f];
		}
		exflib_float_mul(h, cs, S+j*s64, pr64);
		exflib_float_mul(t, sn, e+j*s64, pr64);
		exflib_float_add(h, h, t, pr64);

		exflib_float_mul(e+j*s64, cs, e+j*s64, pr64);
		exflib_float_mul(tmp, sn, S+j*s64, pr64);
		exflib_float_sub(e+j*s64, e+j*s64, tmp, pr64);

		exflib_float_mul(g, sn, S+(j+1)*s64, pr64);
		exflib_float_mul(S+(j+1)*s64, cs, S+(j+1)*s64, pr64);

		if(wantv) {
		    for(i = 0; i < col; i++) {

			exflib_float_mul(t, cs, V+(i*col+j)*s64, pr64);
			exflib_float_mul(tmp, sn, V+(i*col+j+1)*s64, pr64);
			exflib_float_add(t, t, tmp, pr64);

			exflib_float_mul(V+(i*col+j+1)*s64, cs, V+(i*col+j+1)*s64, pr64);
			exflib_float_mul(tmp, sn, V+(i*col+j)*s64, pr64);
			exflib_float_sub(V+(i*col+j+1)*s64, V+(i*col+j+1)*s64, tmp, pr64);
			for(f = 0; f < s64; f++)
			    V[(i*col+j)*s64+f] = t[f];
		    }
		}
		exflib_mex_hypot(t, h, g, pr64);
		exflib_float_div(cs, h, t, pr64);
		exflib_float_div(sn, g, t, pr64);
		for(f = 0; f < s64; f++)
		    S[j*s64+f] = t[f];

		exflib_float_mul(h, cs, e+j*s64, pr64);
		exflib_float_mul(tmp, sn, S+(j+1)*s64, pr64);
		exflib_float_add(h, h, tmp, pr64);

		exflib_float_mul(S+(j+1)*s64, S+(j+1)*s64, cs, pr64);
		exflib_float_mul(tmp, sn, e+j*s64, pr64);
		exflib_float_sub(S+(j+1)*s64, S+(j+1)*s64, tmp, pr64);

		exflib_float_mul(g, sn, e+(j+1)*s64, pr64);
		exflib_float_mul(e+(j+1)*s64, cs, e+(j+1)*s64, pr64);

		if(wantu && (j < row-1)) {
		    for(i = 0; i < row; i++) {
			exflib_float_mul(t, cs, U+(i*nu+j)*s64, pr64);
			exflib_float_mul(tmp, sn, U+(i*nu+j+1)*s64, pr64);
			exflib_float_add(t, t, tmp, pr64);

			exflib_float_mul(U+(i*nu+j+1)*s64, U+(i*nu+j+1)*s64, cs, pr64);
			exflib_float_mul(tmp, sn, U+(i*nu+j)*s64, pr64);
			exflib_float_sub(U+(i*nu+j+1)*s64, U+(i*nu+j+1)*s64, tmp, pr64);

			for(f = 0; f < s64; f++)
			    U[(i*nu+j)*s64+f] = t[f];
		    }
		}
	    }
	    for(f = 0; f < s64; f++)
		e[(p-2)*s64+f] = h[f];
	    iter = iter + 1;

#ifdef __cplusplus
	    delete [] work3;
#else
	    free( work3 );
#endif	    
	}
            break;

            /* Convergence. */

	case 4: {

	    /* Make the singular values positive. */
   
	    if( ((S[k*s64] & EXFLIB_MASK_SGN) != 0) || (S[k*s64] == 0) ) {
		if( (S[k*s64] & EXFLIB_MASK_SGN) != 0){
		    S[k*s64] ^= EXFLIB_MASK_SGN;
		} else {
		    for(f = 0; f < s64; f++)
			S[k*s64+f] = 0;
		}
		if(wantv) {
		    for(i = 0; i <= pp; i++) {
                        V[(i*col+k)*s64] ^= EXFLIB_MASK_SGN;
		    }
		}
	    }
   
	    /* Order the singular values. */
   
	    while(k < pp) {
		if(!exflib_float_lt(S+k*s64, S+(k+1)*s64, pr64)) {
		    break;
		}
		
		for(f = 0; f < s64; f++)
		    t[f] = S[k*s64+f];
		for(f = 0; f < s64; f++)
		    S[k*s64+f] = S[(k+1)*s64+f];
		for(f = 0; f < s64; f++)
		    S[(k+1)*s64+f] = t[f];

		if(wantv && (k < col-1)) {
		    for(i = 0; i < col; i++) {
			for(f = 0; f < s64; f++)
			    t[f] = V[(i*col+k+1)*s64+f];
			for(f = 0; f < s64; f++)
			    V[(i*col+k+1)*s64+f] = V[(i*col+k)*s64+f];
			for(f = 0; f < s64; f++)
			    V[(i*col+k)*s64+f] = t[f];
		    }
		}
		if(wantu && (k < row-1)) {
		    for(i = 0; i < row; i++) {
			for(f = 0; f < s64; f++)
			    t[f] = U[(i*nu+k+1)*s64+f];
			for(f = 0; f < s64; f++)
			    U[(i*nu+k+1)*s64+f] = U[(i*nu+k)*s64+f];
			for(f = 0; f < s64; f++)
			    U[(i*nu+k)*s64+f] = t[f];
		    }
		}
		k++;
	    }
	    iter = 0;
	    p--;
	}
            break;
	}
    }


#ifdef __cplusplus
    delete [] e;
    delete [] work;
    delete [] work_sqrt;
    delete [] tmp;
#else
    free( e );
    free( work );
    free( work_sqrt );
    free( tmp );
#endif

    return;
}

/* End of file */
